// CodeMirror, copyright (c) by Marijn Haverbeke and others
// Distributed under an MIT license: https://codemirror.net/5/LICENSE

(function(mod) {
  if (typeof exports == "object" && typeof module == "object") // CommonJS
    mod(require("../../lib/codemirror"));
  else if (typeof define == "function" && define.amd) // AMD
    define(["../../lib/codemirror"], mod);
  else // Plain browser env
    mod(CodeMirror);
})(function(CodeMirror) {
"use strict";

CodeMirror.defineMode("b4x", function(conf, parserConf) {
    var ERRORCLASS = 'error';

    function wordRegexp(words) {
        return new RegExp("^((" + words.join(")|(") + "))\\b", "i");
    }

    var singleOperators = new RegExp("^[\\+\\-\\*/%&\\\\|\\^~<>!]");
    var singleDelimiters = new RegExp('^[\\(\\)\\[\\]\\{\\}@,:`=;\\.]');
    var doubleOperators = new RegExp("^((==)|(<>)|(<=)|(>=)|(<>)|(<<)|(>>)|(//)|(\\*\\*))");
    var doubleDelimiters = new RegExp("^((\\+=)|(\\-=)|(\\*=)|(%=)|(/=)|(&=)|(\\|=)|(\\^=))");
    var tripleDelimiters = new RegExp("^((//=)|(>>=)|(<<=)|(\\*\\*=))");
    var identifiers = new RegExp("^[_A-Za-z][_A-Za-z0-9]*");

    var openingKeywords = ['class','module', 'sub','enum','select','do while','if','function', 'get','set','property', 'try', 'structure', 'synclock', 'using', 'with','for','for each', 'iif', 'do until', '$"'];
    var middleKeywords = ['then','else','else if','case','case else', 'catch', 'finally', 'step', 'to', 'exit','as'];
    var endKeywords = ['next','loop','end if','end select','end try', '"$'];

    var operatorKeywords = ['and', 'or', 'xor', 'in', 'not', 'is', 'isnot', 'like', 'mod','power'];
    var wordOperators = wordRegexp(operatorKeywords);

    var commonKeywords = ["abs","acos","acosd","array","as","array as", "asc","asin","asind","atan","atan2","atan2d","atand","bytestostring","callsub","callsub2","callsub3","callsubdelayed","callsubdelayed2","callsubdelayed3","catch","ce","ceil","charstostring","chr","continue","cos","cosd","cpi","createmap","crlf","dim","do","exit","false","floor","gettype","if","is","isnumber","loadbitmap","loadbitmapresize","loadbitmapsample","log","logarithm","logcolor","max","me","min","not","numberformat","numberformat2","power","quote","regex","return","rnd","rndseed","round","round2","select","sender","sin","sind","sleep","smartstringformatter","sqrt","sub","subexists","tab","tan","tand","true","try","type","until","wait for","while", "#const", "#else", "#elseif", "#end", "#if", "#region", "#end region","private sub", "public sub","end sub", "button_click()", "const", "type", "public", "private", "file", "wait for" ];

    var commontypes = ['object', 'boolean', 'char', 'string', 'byte', 'short', 'int', 'uint', 'integer', 'long', 'ulong', 'double', 'float', 'date', 'datetime', "javaobject","accordion","animatedcounter","anotherprogressbar","b4xbreadcrumb","b4xcombobox","b4xfloattextfield","b4ximageview","b4xloadingindicator","b4xplusminus","b4xradiobutton","b4xseekbar","b4xswitch","button","canvas","checkbox","choicebox","colorpicker","combobox","customlistview","datepicker","htmleditor","label","listview","madewithlove","menubar","pagination","pane","progressbar","progressindicator","radiobutton","roundslider","scrollinglabel","scrollpane","slider","spinner","splitpane","swiftbutton","tableview","tabpane","textarea","textfield","togglebutton","treetableview","treeview","webview", "list", "map", "mouseevent", "b4xview", "bitmap", "form", "resumablesub", "null", "activity", "edittext", "horizontalscrollview", "imageview","panel", "scrollview", "seekbar", "tabhost", "stringbuilder"];

	var propertieswords = [".text",".tag",".left",".top",".width",".height",".size", ".keys", ".values", ""];
	
	var methodswords = [".initialize",".as()","_click", "_longclick", "_touch", "_down", "_up", "_keypress", "_keyup", "_itemclick", "_itemlongclick", "_checkedchange", "_enterpressed", "_focuschanged", "_textchanged", "_scrollchanged", "_valuechanged", "_tabchanged", "_overrideurl", "_pagefinished", "_action", "_focuschanged", "_mouseclicked", "_mousedragged", "_mousemoved", "_mouseentered", "_mouseexited", "_mousepressed", "_mousereleased", "_resize", "_selectedindexchanged", "_valueschange", "_selectedchange", "_hscrollchanged", "_vscrollchanged", ".clickcount", ".consume", ".middlebuttondown", ".middlebuttonpressed", ".primarybuttondown", ".primarybuttonpressed", ".secondarybuttondown", ".secondarybuttonpressed", ".x", ".y", ".append()", ".insert()", ".isinitialized", ".length", ".remove()", ".tostring", ".openoutput()", ".openinput()", ".writestring()", ".readstring()", ".writelist()", ".readlist()", ".writemap()", ".readmap()", ".writebytes()", ".readbytes()", ".copy()", ".copy2()", ".delete()", ".listfiles()", ".size()", ".makedir()", ".initialize2()", ".add()", ".addall()", "addallat()", ".insertat()", ".removeat()", ".clear", ".sort()", ".get()", ".set()", ".put()", ".getkeyat()", ".getvalueat()", ".containskey()", ".removekey()",".charat()",".compareto()",".contains()",".endswith()",".equalsingnorecase()",".getbytes()",".indexof()",".indexof2()",".lastindexof()",".lastindexof2()",".length()",".replace()",".startswith()",".substring()",".substring2()",".tolowercase",".touppercase",".trim", "" ];

    var keywords = wordRegexp(commonKeywords);
    var types = wordRegexp(commontypes);
    var stringPrefixes = '"';

    var opening = wordRegexp(openingKeywords);
    var middle = wordRegexp(middleKeywords);
    var closing = wordRegexp(endKeywords);
    var doubleClosing = wordRegexp(['end']);
    var doOpening = wordRegexp(['do']);

    var indentInfo = null;

    CodeMirror.registerHelper("hintWords", "b4x", openingKeywords.concat(middleKeywords).concat(endKeywords)
                                .concat(operatorKeywords).concat(commonKeywords).concat(commontypes).concat(propertieswords).concat(methodswords));

    function indent(_stream, state) {
      state.currentIndent++;
    }

    function dedent(_stream, state) {
      state.currentIndent--;
    }
    // tokenizers
    function tokenBase(stream, state) {
        if (stream.eatSpace()) {
            return null;
        }

        var ch = stream.peek();

        // Handle Comments
        if (ch === "'") {
            stream.skipToEnd();
            return 'comment';
        }


        // Handle Number Literals
        if (stream.match(/^((&H)|(&O))?[0-9\.a-f]/i, false)) {
            var floatLiteral = false;
            // Floats
            if (stream.match(/^\d*\.\d+F?/i)) { floatLiteral = true; }
            else if (stream.match(/^\d+\.\d*F?/)) { floatLiteral = true; }
            else if (stream.match(/^\.\d+F?/)) { floatLiteral = true; }

            if (floatLiteral) {
                // Float literals may be "imaginary"
                stream.eat(/J/i);
                return 'number';
            }
            // Integers
            var intLiteral = false;
            // Hex
            if (stream.match(/^&H[0-9a-f]+/i)) { intLiteral = true; }
            // Octal
            else if (stream.match(/^&O[0-7]+/i)) { intLiteral = true; }
            // Decimal
            else if (stream.match(/^[1-9]\d*F?/)) {
                // Decimal literals may be "imaginary"
                stream.eat(/J/i);
                // TODO - Can you have imaginary longs?
                intLiteral = true;
            }
            // Zero by itself with no other piece of number.
            else if (stream.match(/^0(?![\dx])/i)) { intLiteral = true; }
            if (intLiteral) {
                // Integer literals may be "long"
                stream.eat(/L/i);
                return 'number';
            }
        }

        // Handle Strings
        if (stream.match(stringPrefixes)) {
            state.tokenize = tokenStringFactory(stream.current());
            return state.tokenize(stream, state);
        }

        // Handle operators and Delimiters
        if (stream.match(tripleDelimiters) || stream.match(doubleDelimiters)) {
            return null;
        }
        if (stream.match(doubleOperators)
            || stream.match(singleOperators)
            || stream.match(wordOperators)) {
            return 'operator';
        }
        if (stream.match(singleDelimiters)) {
            return null;
        }
        if (stream.match(doOpening)) {
            indent(stream,state);
            state.doInCurrentLine = true;
            return 'keyword';
        }
        if (stream.match(opening)) {
            if (! state.doInCurrentLine)
              indent(stream,state);
            else
              state.doInCurrentLine = false;
            return 'keyword';
        }
        if (stream.match(middle)) {
            return 'keyword';
        }

        if (stream.match(doubleClosing)) {
            dedent(stream,state);
            dedent(stream,state);
            return 'keyword';
        }
        if (stream.match(closing)) {
            dedent(stream,state);
            return 'keyword';
        }

        if (stream.match(types)) {
            return 'keyword';
        }

        if (stream.match(keywords)) {
            return 'keyword';
        }

        if (stream.match(identifiers)) {
            return 'variable';
        }

        // Handle non-detected items
        stream.next();
        return ERRORCLASS;
    }

    function tokenStringFactory(delimiter) {
        var singleline = delimiter.length == 1;
        var OUTCLASS = 'string';

        return function(stream, state) {
            while (!stream.eol()) {
                stream.eatWhile(/[^'"]/);
                if (stream.match(delimiter)) {
                    state.tokenize = tokenBase;
                    return OUTCLASS;
                } else {
                    stream.eat(/['"]/);
                }
            }
            if (singleline) {
                if (parserConf.singleLineStringErrors) {
                    return ERRORCLASS;
                } else {
                    state.tokenize = tokenBase;
                }
            }
            return OUTCLASS;
        };
    }


    function tokenLexer(stream, state) {
        var style = state.tokenize(stream, state);
        var current = stream.current();

        // Handle '.' connected identifiers
        if (current === '.') {
            style = state.tokenize(stream, state);
            if (style === 'variable') {
                return 'variable';
            } else {
                return ERRORCLASS;
            }
        }


        var delimiter_index = '[({'.indexOf(current);
        if (delimiter_index !== -1) {
            indent(stream, state );
        }
        if (indentInfo === 'dedent') {
            if (dedent(stream, state)) {
                return ERRORCLASS;
            }
        }
        delimiter_index = '])}'.indexOf(current);
        if (delimiter_index !== -1) {
            if (dedent(stream, state)) {
                return ERRORCLASS;
            }
        }

        return style;
    }

    var external = {
        electricChars:"",
        startState: function() {
            return {
              tokenize: tokenBase,
              lastToken: null,
              currentIndent: 0,
              nextLineIndent: 0,
              doInCurrentLine: false


          };
        },

        token: function(stream, state) {
            if (stream.sol()) {
              state.currentIndent += state.nextLineIndent;
              state.nextLineIndent = 0;
              state.doInCurrentLine = 0;
            }
            var style = tokenLexer(stream, state);

            state.lastToken = {style:style, content: stream.current()};



            return style;
        },

        indent: function(state, textAfter) {
            var trueText = textAfter;		/* textAfter.replace(/^\s+|\s+$/g, '') ; */
            if (trueText.match(closing) || trueText.match(doubleClosing) || trueText.match(middle)) return conf.indentUnit*(state.currentIndent-1);
            if(state.currentIndent < 0) return 0;
			state.currentIndent = state.currentIndent + 1;
            return (state.currentIndent) * conf.indentUnit;
        },

        lineComment: "'"
    };
    return external;
});

CodeMirror.defineMIME("text/x-vb", "b4x");

});
